==> amqp10-2.0.4.tgz: package/lib/link.js <==
'use strict';

var EventEmitter = require('events').EventEmitter,
    Promise = require('bluebird'),
    util = require('util'),

    StateMachine = require('stately.js'),

    debug = require('debug')('amqp10:link'),
    u = require('./utilities'),

    AttachFrame = require('./frames/attach_frame'),
    DetachFrame = require('./frames/detach_frame');

function Link(session, handle, linkPolicy) {
  this.policy = linkPolicy;
  this.session = session;
  this.handle = handle;
  this.remoteHandle = undefined;

  this._onAttach = [];
  if (this.policy && this.policy.reattach) {
    this._timeouts = u.generateTimeouts(this.policy.reattach);
  }

  var self = this;
  var stateMachine = {
    'DETACHED': {
      sendAttach: function() {
        return this.ATTACHING;
      }
    },
    'ATTACHING': {
      attachReceived: function() {
        return this.ATTACHED;
      }
    },
    'ATTACHED': {
      sendDetach: function() {
        return this.DETACHING;
      },
      detachReceived: function() {
        self._sendDetach();
        return this.DETACHING;
      },
      forceDetach: function() {
        return this.DETACHED;
      }
    },
    'DETACHING': {
      detachReceived: function() {
        return this.DETACHED;
      },
      detached: function() {
        return this.DETACHED;
      }
    }
  };

  this.linkSM = new StateMachine(stateMachine).bind(function(event, oldState, newState) {
    debug(self.name + ':' + self.handle + ': Transitioning from ' + oldState + ' to ' + newState + ' due to ' + event);
  });
}

util.inherits(Link, EventEmitter);

// On receipt of a message.  Message payload given as argument.
Link.MessageReceived = 'message';

// Since 'error' events are "special" in Node (as in halt-the-process special),
// using a custom event for errors we receive from the other endpoint. Provides
// received AMQPError as an argument.
Link.ErrorReceived = 'errorReceived';

// On link credit changed.
Link.CreditChange = 'creditChange';

// On completion of detach.
Link.Attached = 'attached';

// On completion of detach.
Link.Detached = 'detached';

// public api
Link.prototype.state = function() {
  return this.linkSM.getMachineState().toLowerCase();
};

Link.prototype.attach = function() {
  this.linkSM.sendAttach();
  var attachFrame = new AttachFrame(this.policy.attach);
  attachFrame.channel = this.session.channel;
  debug('attach CH=' + attachFrame.channel + ', Handle=' + attachFrame.handle);

  this.name = attachFrame.name;
  this.role = attachFrame.role;
  this.linkCredit = 0;
  this.totalCredits = 0;
  this.available = 0;
  this.drain = false;
  this.session.connection.sendFrame(attachFrame);
};

Link.prototype.detach = function() {
  var self = this;
  this._timeouts = undefined; // Disable any re-attachment policy.
  var detachPromise = new Promise(function(resolve, reject) {
    var onError = function(err) { reject(err); };
    self.once(Link.ErrorReceived, onError);
    self.once(Link.Detached, function(info) {
      self.removeListener(Link.ErrorReceived, onError);
      if (!!info.error) return reject(info.error);
      if (!info.closed) return reject('link not closed');
      resolve();
    });
  });

  this.linkSM.sendDetach();
  this._sendDetach();
  return detachPromise;
};

///
/// Force link state to detached without sending detach message - usually due to forcible disconnect or unmap from above.
/// Important bit is that this should not trigger auto-reattach behavior as that'll happen with reconnect.
///
Link.prototype.forceDetach = function() {
  debug('Force-detach for ' + this.name + '. Current state: ' + this.state());
  this.linkSM.forceDetach();
};

// private api
Link.prototype._resolveAttachPromises = function(err, link) {
  while (this._onAttach.length) {
    var attachPromise = this._onAttach.shift();
    attachPromise(err, link);
  }
};

Link.prototype._attachReceived = function(attachFrame) {
  this.linkSM.attachReceived();

  // process params.
  this.remoteHandle = attachFrame.handle;
  this.session._linksByRemoteHandle[this.remoteHandle] = this;
  debug(this.name + ': attached CH=[' + this.session.channel + '=>' + attachFrame.channel + '], Handle=[' + this.handle + '=>' + attachFrame.handle + ']');

  this.emit(Link.Attached, this);
  this._resolveAttachPromises(null, this);

  this._checkCredit();
};

// default implementation does nothing
Link.prototype._checkCredit = function() {};

Link.prototype._detachReceived = function(frame) {
  this.linkSM.detachReceived();
  if (this.linkSM.getMachineState() === 'DETACHING') this.linkSM.detached();
  this._detached(frame);
};

Link.prototype._sendDetach = function() {
  var detachFrame = new DetachFrame({
    handle: this.handle,
    channel: this.session.channel,
    closed: true
  });

  this.session.connection.sendFrame(detachFrame);
};

Link.prototype._detached = function(frame) {
  if (frame && frame.error) {
    this.emit(Link.ErrorReceived, frame.error);
  }

  if (this.remoteHandle !== undefined) {
    delete this.session._linksByRemoteHandle[this.remoteHandle];
    this.remoteHandle = undefined;
  }

  this.emit(Link.Detached, { closed: frame.closed, error: frame.error });
  this._resolveAttachPromises(frame.error ? frame.error : 'link closed');

  var self = this;
  if (!self.shouldReattach()) return;

  if (!self._timeouts.length) self._timeouts = u.generateTimeouts(self.policy.reattach);

  setTimeout(function() {
    self._attemptReattach();
  }, self._timeouts.shift());
};

Link.prototype.shouldReattach = function() {
  if (!this.session || !this._timeouts) return false;
  if (!this._timeouts.length && !this.policy.reattach.forever) return false;
  return true;
};

Link.prototype._attemptReattach = function() {
  debug('Attempting to reattach ' + this.name);
  this.attach();
};

module.exports = Link;

==> amqp10-2.1.3.tgz: package/lib/link.js <==
'use strict';

var EventEmitter = require('events').EventEmitter,
    Promise = require('bluebird'),
    util = require('util'),
    errors = require('./errors'),

    StateMachine = require('stately.js'),

    debug = require('debug')('amqp10:link'),
    u = require('./utilities'),

    AttachFrame = require('./frames/attach_frame'),
    DetachFrame = require('./frames/detach_frame');

function Link(session, handle, linkPolicy) {
  this.policy = linkPolicy;
  this.session = session;
  this.handle = handle;
  this.remoteHandle = undefined;

  this._onAttach = [];
  if (this.policy && this.policy.reattach) {
    this._timeouts = u.generateTimeouts(this.policy.reattach);
  }

  var self = this;
  var stateMachine = {
    'DETACHED': {
      sendAttach: function() {
        return this.ATTACHING;
      }
    },
    'ATTACHING': {
      attachReceived: function() {
        return this.ATTACHED;
      }
    },
    'ATTACHED': {
      sendDetach: function() {
        return this.DETACHING;
      },
      detachReceived: function() {
        self._sendDetach();
        return this.DETACHING;
      },
      forceDetach: function() {
        return this.DETACHED;
      }
    },
    'DETACHING': {
      detachReceived: function() {
        return this.DETACHED;
      },
      detached: function() {
        return this.DETACHED;
      }
    }
  };

  this.linkSM = new StateMachine(stateMachine).bind(function(event, oldState, newState) {
    debug(self.name + ':' + self.handle + ': Transitioning from ' + oldState + ' to ' + newState + ' due to ' + event);
  });
}

util.inherits(Link, EventEmitter);

// On receipt of a message.  Message payload given as argument.
Link.MessageReceived = 'message';

// Since 'error' events are "special" in Node (as in halt-the-process special),
// using a custom event for errors we receive from the other endpoint. Provides
// received AMQPError as an argument.
Link.ErrorReceived = 'errorReceived';

// On link credit changed.
Link.CreditChange = 'creditChange';

// On completion of detach.
Link.Attached = 'attached';

// On completion of detach.
Link.Detached = 'detached';

// public api
Link.prototype.state = function() {
  return this.linkSM.getMachineState().toLowerCase();
};

Link.prototype.attach = function() {
  this.linkSM.sendAttach();
  var attachFrame = new AttachFrame(this.policy.attach);
  attachFrame.channel = this.session.channel;
  debug('attach CH=' + attachFrame.channel + ', Handle=' + attachFrame.handle);

  this.name = attachFrame.name;
  this.role = attachFrame.role;
  this.linkCredit = 0;
  this.totalCredits = 0;
  this.available = 0;
  this.drain = false;
  this.session.connection.sendFrame(attachFrame);
};

Link.prototype.detach = function() {
  var self = this;
  this._timeouts = undefined; // Disable any re-attachment policy.
  var detachPromise = new Promise(function(resolve, reject) {
    var onError = function(err) { reject(err); };
    self.once(Link.ErrorReceived, onError);
    self.once(Link.Detached, function(info) {
      self.removeListener(Link.ErrorReceived, onError);
      if (!!info.error) return reject(info.error);
      if (!info.closed) return reject('link not closed');
      resolve();
    });
  });

  this.linkSM.sendDetach();
  this._sendDetach();
  return detachPromise;
};

///
/// Force link state to detached without sending detach message - usually due to forcible disconnect or unmap from above.
/// Important bit is that this should not trigger auto-reattach behavior as that'll happen with reconnect.
///
Link.prototype.forceDetach = function() {
  debug('Force-detach for ' + this.name + '. Current state: ' + this.state());
  this.linkSM.forceDetach();
};

// private api
Link.prototype._resolveAttachPromises = function(err, link) {
  while (this._onAttach.length) {
    var attachPromise = this._onAttach.shift();
    attachPromise(err, link);
  }
};

Link.prototype._attachReceived = function(attachFrame) {
  this.linkSM.attachReceived();

  // process params.
  this.remoteHandle = attachFrame.handle;
  this.session._linksByRemoteHandle[this.remoteHandle] = this;
  debug(this.name + ': attached CH=[' + this.session.channel + '=>' + attachFrame.channel + '], Handle=[' + this.handle + '=>' + attachFrame.handle + ']');

  this.emit(Link.Attached, this);
  this._resolveAttachPromises(null, this);

  this._checkCredit();
};

// default implementation does nothing
Link.prototype._checkCredit = function() {};

Link.prototype._detachReceived = function(frame) {
  this.linkSM.detachReceived();
  if (this.linkSM.getMachineState() === 'DETACHING') this.linkSM.detached();
  this._detached(frame);
};

Link.prototype._sendDetach = function() {
  var detachFrame = new DetachFrame({
    handle: this.handle,
    channel: this.session.channel,
    closed: true
  });

  this.session.connection.sendFrame(detachFrame);
};

Link.prototype._detached = function(frame) {
  if (frame && frame.error) {
    this.emit(Link.ErrorReceived, errors.wrapProtocolError(frame.error));
  }

  if (this.remoteHandle !== undefined) {
    delete this.session._linksByRemoteHandle[this.remoteHandle];
    this.remoteHandle = undefined;
  }

  this.emit(Link.Detached, { closed: frame.closed, error: errors.wrapProtocolError(frame.error) });
  this._resolveAttachPromises(frame.error ? frame.error : 'link closed');

  var self = this;
  if (!self.shouldReattach()) return;

  if (!self._timeouts.length) self._timeouts = u.generateTimeouts(self.policy.reattach);

  setTimeout(function() {
    self._attemptReattach();
  }, self._timeouts.shift());
};

Link.prototype.shouldReattach = function() {
  if (!this.session || !this._timeouts) return false;
  if (!this._timeouts.length && !this.policy.reattach.forever) return false;
  return true;
};

Link.prototype._attemptReattach = function() {
  debug('Attempting to reattach ' + this.name);
  this.attach();
};

module.exports = Link;
